import abc


class AInsertFunction(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def build_functions(self) -> list:
        raise NotImplementedError

    def build_view(self) -> list:
        return list()

    def insert_functions(self, conn, do_commit: bool = True):
        cursor = conn.cursor()
        for view in self.build_view():
            cursor.execute(view)
        for function in self.build_functions():
            cursor.execute(function)
        if do_commit:
            conn.commit()



def create_enum_type(name: str, enum, conn):
    cursor = conn.cursor()
    enum_lst = list()
    for n in all_enum_values(enum):
        enum_lst.append(f"'{n}'")
    sql = f"DROP TYPE IF EXISTS {name} CASCADE; " \
          f"CREATE TYPE {name} AS ENUM ({','.join(enum_lst)});"
    # print(sql)
    cursor.execute(sql)
    # print(f"Inserted enum type [{name}].")


def all_enum_values(cls) -> list:
    role_names = [member.value for role, member in cls.__members__.items()]
    return role_names


def all_enum_names(cls) -> list:
    val_list = all_enum_values(cls)
    res = list()
    for v in val_list:
        res.append(cls(v).name)
    return res



from volsite_postgres_common.db.CC import CC
from volsite_postgres_common.table.description import insert_empty_description
from volsite_postgres_common.table.series import set_max_id_for_serial_column


# @note PostgreSQL treats NULL as distinct value
def create_table_descriptions(conn, sc: str, table: str):
    sc_table = f'"{sc}"."{table}"'
    cursor = conn.cursor()
    cursor.execute(
        f" CREATE TABLE IF NOT EXISTS "
        f" {sc_table} "
        f"({CC.description_id}        BIGSERIAL PRIMARY KEY, "
        f" {CC.formatted_description} TEXT NOT NULL, "
        f" {CC.plain_description}     TEXT NOT NULL, "
        f" {CC.length}                INTEGER NOT NULL, "

        f" {CC.used_count}     INTEGER   NOT NULL DEFAULT 1, "
        f" {CC.built_time}     TIMESTAMP NOT NULL, "
        f" {CC.last_used_time} TIMESTAMP NOT NULL "
        f" );")

    cursor.execute(
        f" CREATE UNIQUE INDEX {sc}_{table}_{CC.formatted_description} "
        f" ON {sc_table}({CC.formatted_description});")

    insert_empty_description(sc_table, CC.description_id, cursor)
    set_max_id_for_serial_column(sc, table, CC.description_id, cursor)
    conn.commit()



from psycopg2._json import Json
from typing import Final
import json
from volsite_postgres_common.api.CA import CA
from volsite_postgres_common.db.BFn import BFn
from volsite_postgres_common.db.CFn import CFn
from volsite_postgres_common.test.Timer import Timer
from volsite_postgres_common.test.db.ATestDb import ATestDb


def json_fn(fn: str, input_j: dict, conn, attList,
            do_commit: bool = False,
            print_input_output: bool = True,
            print_long_att: bool = True):  # -> dict:
    cursor = conn.cursor()
    if print_input_output:
        print(f'==== [FN] {fn} ====')
        print('=== Input ===')
        print('<code>')
        print_json_by_attributes(input_j, attList, print_long_att)
        print('</code>')
    cursor.execute(f'SELECT {fn}( %s::JSONB ) AS {CA.Result}', (
        Json(input_j),
    ))
    rows = cursor.fetchall()
    assert 1 == len(rows)
    # print('[json_fn] rows[0] = %r' % rows[0])
    if do_commit:
        conn.commit()
    output = rows[0][CA.Result]
    if print_input_output:
        print('=== Output ===')
        print('<code>')
        print_json_by_attributes(output, attList, print_long_att)
        print('</code>')
    return output


def print_json_by_attributes(j, attList, print_long_att: bool = False):
    if not print_long_att:
        print(json.dumps(j, indent=4, sort_keys=True))
        return

    abb_att = {}

    for a in attList:
        for name in a.__dict__:
            abb = a.__dict__[name]
            abb_att[f"\"{abb}\":"] = f"\"{abb}__{name}\":"

    res = json.dumps(j, indent=4, sort_keys=True)
    for att in abb_att.keys():
        res = res.replace(att, abb_att[att])
    print(res)


def json_fn_db(
        fn: str, input_j: dict,
        test_db: ATestDb, attList,
        do_commit: bool = False,
        print_long_att: bool = True):  # -> dict:
    with Timer(fn):
        return json_fn(fn, input_j, test_db.p_conn, attList,
                       do_commit=do_commit,
                       print_input_output=test_db.print_input_output,
                       print_long_att=print_long_att)


fn_jsonb_array_2_int_array: Final = (
    f" CREATE OR REPLACE FUNCTION "
    f" {CFn.jsonb_array_2_int_array} (JSONB) "
    f" RETURNS INT[] "
    f" AS $$ "
    f"   SELECT array_agg(x)::INT[] || ARRAY[]::INT[] FROM {BFn.jsonb_array_elements_text}($1) t(x);"
    f" $$ "
    f" LANGUAGE SQL "
    f" IMMUTABLE;")

fn_jsonb_array_2_smallint_array: Final = (
    f" CREATE OR REPLACE FUNCTION "
    f" {CFn.jsonb_array_2_smallint_array} (JSONB) "
    f" RETURNS SMALLINT[] "
    f" AS $$ "
    f"   SELECT array_agg(x)::SMALLINT[] || ARRAY[]::SMALLINT[] FROM {BFn.jsonb_array_elements_text}($1) t(x);"
    f" $$ "
    f" LANGUAGE SQL "
    f" IMMUTABLE;")

fn_jsonb_array_2_bigint_array: Final = (
    f" CREATE OR REPLACE FUNCTION "
    f" {CFn.jsonb_array_2_bigint_array} (JSONB) "
    f" RETURNS BIGINT[] "
    f" AS $$ "
    f"   SELECT array_agg(x)::BIGINT[] || ARRAY[]::BIGINT[] FROM {BFn.jsonb_array_elements_text}($1) t(x);"
    f" $$ "
    f" LANGUAGE SQL "
    f" IMMUTABLE;")

fn_jsonb_array_2_text_array: Final = (
    f" CREATE OR REPLACE FUNCTION "
    f" {CFn.jsonb_array_2_text_array} (JSONB) "
    f" RETURNS TEXT[] "
    f" AS $$ "
    f"   SELECT array_agg(x)::TEXT[] || ARRAY[]::TEXT[] FROM {BFn.jsonb_array_elements_text}($1) t(x);"
    f" $$ "
    f" LANGUAGE SQL "
    f" IMMUTABLE;")

fn_jsonb_remove_key: Final = (
    f" CREATE OR REPLACE FUNCTION "
    f" {CFn.jsonb_remove_key} (JSONB, TEXT) "
    f" RETURNS JSONB "
    f" AS $$ "
    f"   SELECT $1 - $2;"
    f" $$ "
    f" LANGUAGE SQL "
    f" IMMUTABLE;")

# @ref https://stackoverflow.com/questions/36171737/how-to-count-setof-number-of-keys-of-json-in-postgresql
fn_count_jsonb_keys: Final = (
    f" CREATE OR REPLACE FUNCTION "
    f" {CFn.count_jsonb_keys} (JSONB) "
    f" RETURNS BIGINT "
    f" AS $$ "
    f"   SELECT COUNT(*) FROM (SELECT {BFn.jsonb_object_keys}($1)) v;"
    f" $$ "
    f" LANGUAGE SQL "
    f" IMMUTABLE;")
'''

fn_jsonb_add_key_value: Final = (
    f" CREATE OR REPLACE FUNCTION "
    f" {Fn.jsonb_add_key_value} (JSONB, JSONB) "
    f" RETURNS JSONB "
    f" AS $$ "
    f"   SELECT $1 || $2;"
    f" $$ "
    f" LANGUAGE SQL "
    f" IMMUTABLE;")
'''

# @ref https://stackoverflow.com/questions/47064098/postgresql-add-object-to-jsonb-array-by-function
fn_jsonb_add_key_val_elements: Final = (
    f" CREATE OR REPLACE FUNCTION "
    f" {CFn.jsonb_add_key_value_elements} (_arr JSONB, _val JSONB) "
    f" RETURNS JSONB "
    f" AS $$ "
    f"   SELECT jsonb_agg(elem || _val)"
    f"      FROM jsonb_array_elements(_arr) elem"
    f" $$ "
    f" LANGUAGE SQL "
    f" IMMUTABLE;")


def insert_util_fn__json(conn):
    cursor = conn.cursor()
    cursor.execute(fn_jsonb_array_2_bigint_array)
    cursor.execute(fn_jsonb_array_2_int_array)
    cursor.execute(fn_jsonb_array_2_smallint_array)
    cursor.execute(fn_jsonb_array_2_text_array)
    cursor.execute(fn_jsonb_remove_key)
    cursor.execute(fn_count_jsonb_keys)
    # cursor.execute(fn_jsonb_add_key_value)
    cursor.execute(fn_jsonb_add_key_val_elements)
    conn.commit()



import abc
from typing import Final

from volsite_postgres_common.db.FnT import FnT


class Arg:
    input: Final[str] = "_input"
    result: Final[str] = "_result"

    def __init__(self, name: str, d_type: FnT):
        self.name = name
        self.type = d_type

    def to_str(self) -> str:
        return f"{self.name}\t{self.type.value}"


class APlPgSqlFunction(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def name(self) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def in_args(self) -> list:
        raise NotImplementedError

    @abc.abstractmethod
    def out_args(self) -> list:
        raise NotImplementedError

    def declare(self) -> list:
        return list()

    @abc.abstractmethod
    def body(self) -> str:
        raise NotImplementedError

    def build_function(self) -> str:
        fn = f" CREATE OR REPLACE FUNCTION {self.name()} \n"
        if len(self.in_args()) + len(self.out_args()) > 0:
            fn += " (\n"
            args = list()
            for i in self.in_args():
                args.append(f" IN {i.to_str()}")
            for o in self.out_args():
                args.append(f" OUT {o.to_str()}")
            fn += ",\n".join(args)
            fn += "\n )\n"
        fn += " AS \n"
        fn += " $$ \n"
        if len(self.declare()) > 0:
            fn += " DECLARE \n"
            for d in self.declare():
                fn += f" {d.to_str()};\n"
        fn += " BEGIN \n\n"
        fn += self.body()
        fn += "\n\n END; \n"
        fn += " $$ \n"
        fn += " LANGUAGE PlPgSQL;"

        return fn


class AJsonPlPgSqlFunction(APlPgSqlFunction, metaclass=abc.ABCMeta):
    def in_args(self) -> list:
        return [
            Arg(Arg.input, FnT.JSONB)
        ]

    def out_args(self) -> list:
        return [
            Arg(Arg.result, FnT.JSONB)
        ]


class ASqlFunction(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def name(self) -> str:
        raise NotImplementedError

    @abc.abstractmethod
    def in_args(self) -> list:
        raise NotImplementedError

    @abc.abstractmethod
    def return_type(self) -> FnT:
        raise NotImplementedError

    @abc.abstractmethod
    def body(self) -> str:
        raise NotImplementedError

    def build_function(self) -> str:
        fn = f" CREATE OR REPLACE FUNCTION {self.name()} \n"
        if len(self.in_args()) > 0:
            fn += " (\n"
            args = list()
            for i in self.in_args():
                args.append(f" {i.to_str()}")
            fn += ",\n".join(args)
            fn += "\n )\n"
        if self.return_type() is None:
            fn += " RETURNS VOID\n"
        else:
            fn += f" RETURNS {self.return_type().value} \n"
        fn += " AS \n"
        fn += " $$ \n\n"
        fn += self.body()
        fn += "\n\n $$ \n"
        fn += " LANGUAGE SQL;"

        return fn


class AJsonSqlFunction(ASqlFunction, metaclass=abc.ABCMeta):
    def in_args(self) -> list:
        return [
            Arg(Arg.input, FnT.JSONB)
        ]

    def return_type(self) -> FnT:
        return FnT.JSONB


from aenum import Enum


class FnT(Enum):  # Function Arg Types
    BIGINT = 'BIGINT'
    BIGINT_ARRAY = 'BIGINT[]'

    BOOLEAN = 'BOOLEAN'
    
    FLOAT = 'FLOAT'

    INT = 'INT'
    INT_ARRAY = 'INT[]'

    JSONB = 'JSONB'
    JSONB_ARRAY = 'JSONB[]'

    SMALLINT = 'SMALLINT'
    SMALLINT_ARRAY = 'SMALLINT[]'

    TEXT = 'TEXT'
    TEXT_ARRAY = 'TEXT[]'

    TIMESTAMP = 'TIMESTAMP'

    VOID = 'VOID'



import abc
from abc import ABC
from typing import Final

from volsite_postgres_common.db.CC import CC
from volsite_postgres_common.db.FnT import FnT
from volsite_postgres_common.fn.function import Arg, ASqlFunction


class APFnUpsertDescription(ASqlFunction, ABC):
    _formatted_: Final[str] = '_formatted_'
    _plain_: Final[str] = '_plain_'

    @abc.abstractmethod
    def table_description(self) -> str:  # {ScT.mentor__descriptions}
        raise NotImplementedError

    def in_args(self) -> list:
        return [
            Arg(self._formatted_, FnT.TEXT),
            Arg(self._plain_, FnT.TEXT)
        ]

    def return_type(self) -> FnT:
        return FnT.BIGINT

    def body(self) -> str:
        return (
            f" INSERT INTO "
            f" {self.table_description()} "
            f" ({CC.formatted_description}, "
            f"  {CC.plain_description}, "
            f"  {CC.length}, "
            f"  {CC.built_time}, {CC.last_used_time}) "
            f" VALUES (trim({self._formatted_}), "
            f"         trim({self._plain_}), "
            f"         length(trim({self._plain_})), "
            f" current_timestamp, current_timestamp) "
            f" ON CONFLICT({CC.formatted_description}) "
            f"   DO"
            f"     UPDATE SET "
            f"       {CC.used_count} = {self.table_description()}.{CC.used_count} + 1,  "
            f"       {CC.last_used_time} = current_timestamp"
            f" RETURNING {CC.description_id};"
            )



from typing import Final


# ====== CC: Common Table Column ======
class CC:
    built_time: Final[str] = 'built_time'

    description_id: Final[str] = 'description_id'

    formatted_description: Final[str] = 'formatted_description'

    info: Final = 'info'

    last_used_time: Final[str] = 'last_used_time'
    length: Final[str] = 'length'

    plain_description: Final[str] = 'plain_description'

    text: Final[str] = 'text'

    used_count: Final[str] = 'used_count'



from volsite_postgres_common.api.CA import CA
from volsite_postgres_common.db.CFn import CFn
from volsite_postgres_common.fn.function import Arg


def sql__input_j_2_iid(att: str) -> str:
    return f"({Arg.input}->>'{att}')::TEXT"


def sql__input_j_2_bint_id(att: str) -> str:
    return f"{CFn.id_2_bigint}( ({Arg.input}->>'{att}')::TEXT )"


def sql__input_j_2_int_id(att: str) -> str:
    return f"{CFn.id_2_int}( ({Arg.input}->>'{att}')::TEXT )"


def sql__2_id(*att: str) -> str:
    return f"{CFn.bigint_2_id}({'.'.join(att)})"


def sql__2_bint(*att: str) -> str:
    return f"{CFn.id_2_bigint}({'.'.join(att)})"


def sql__order_by_cols(col_dict: dict, def_col: str) -> str:
    res = f"CASE ({Arg.input}->>'{CA.SortBy}')::TEXT"
    for arg in col_dict.keys():
        res = f"{res}\n          WHEN '{arg}' THEN {col_dict[arg]}"
    res = f"{res}\n          ELSE {def_col}"  # ELSE NULL
    res = f"{res}\nEND"
    return res


def sql__order_by(col_dict: dict, def_col: str) -> str:
    return (
        f" CASE WHEN ({Arg.input}->>'{CA.SortDirection}')::TEXT = '{CA.Ascending}' THEN "
        f"      {sql__order_by_cols(col_dict, def_col)}"
        f" ELSE"
        f"      NULL"
        f" END"
        f" {CA.Ascending},"
        f""
        f" CASE WHEN ({Arg.input}->>'{CA.SortDirection}')::TEXT = '{CA.Descending}' THEN "
        f"      {sql__order_by_cols(col_dict, def_col)}"
        f" ELSE"
        f"      NULL"
        f" END"
        f" {CA.Descending}"
    )



from typing import Final
from random import randrange


# @ref https://stackoverflow.com/questions/9108833/postgres-autoincrement-not-updated-on-explicit-id-inserts
def reset_max_id_for_series(schema: str, table: str, col: str, cursor):
    cursor.execute(f"SELECT setval('{table}_{col}_seq', (SELECT MAX({col}) FROM \"{schema}.{table}\"));")


Min_DB_Id_Series: Final = 1000


# @ref https://www.postgresql.org/docs/current/functions-sequence.html
def set_max_id_for_serial_column(schema: str, table: str, col: str, cursor, min_id=Min_DB_Id_Series):
    cursor.execute(f"SELECT setval(pg_get_serial_sequence('"
                   f"{schema}.{table}', '{col}'), "
                   f"{min_id + randrange(int(min_id/10))}, true);")



from typing import Final

from volsite_postgres_common.db.BFn import BFn
from volsite_postgres_common.db.CFn import CFn
from volsite_postgres_common.fn.insert import insert_function

ID_PREFIX: Final = 'i'
ENUM_PREFIX: Final = 'e'


bigint_2_id: Final = (
    f" CREATE OR REPLACE FUNCTION "
    f" {CFn.bigint_2_id} (_bid BIGINT) "
    f" RETURNS TEXT "
    f" AS"
    f" $$"
    f"   SELECT CONCAT ( '{ID_PREFIX}', _bid)"
    f" $$ "
    f" STRICT"
    f" LANGUAGE SQL "
    f" IMMUTABLE;")

int_2_id: Final = (
    f" CREATE OR REPLACE FUNCTION "
    f" {CFn.int_2_id} (_bid INT) "
    f" RETURNS TEXT "
    f" AS"
    f" $$"
    f"   SELECT CONCAT ( '{ID_PREFIX}', _bid)"
    f" $$ "
    f" STRICT"
    f" LANGUAGE SQL "
    f" IMMUTABLE;")


id_2_bigint: Final = (
    f" CREATE OR REPLACE FUNCTION "
    f" {CFn.id_2_bigint} (_bid TEXT) "
    f" RETURNS BIGINT "
    f" AS"
    f" $$"
    f"   SELECT CAST((SPLIT_PART ( _bid, '{ID_PREFIX}', '2')) AS BIGINT)"
    f" $$ "
    f" STRICT"
    f" LANGUAGE SQL "
    f" IMMUTABLE;")


id_2_int: Final = (
    f" CREATE OR REPLACE FUNCTION "
    f" {CFn.id_2_int} (_bid TEXT) "
    f" RETURNS INT "
    f" AS"
    f" $$"
    f"   SELECT CAST((SPLIT_PART ( _bid, '{ID_PREFIX}', '2')) AS INT)"
    f" $$ "
    f" STRICT"
    f" LANGUAGE SQL "
    f" IMMUTABLE;")

jsonb_id_array_2_bigint_array: Final = (
    f" CREATE OR REPLACE FUNCTION "
    f" {CFn.jsonb_id_array_2_bigint_array} (JSONB) "
    f" RETURNS BIGINT[] "
    f" AS $$ "
    f"   SELECT array_agg({CFn.id_2_bigint}(x))::BIGINT[] || ARRAY[]::BIGINT[] "
    f"      FROM {BFn.jsonb_array_elements_text}($1) t(x);"
    f" $$ "
    f" LANGUAGE SQL "
    f" IMMUTABLE;")


def insert_util_fn__id(conn):
    insert_function(bigint_2_id, CFn.bigint_2_id, conn)
    insert_function(int_2_id, CFn.int_2_id, conn)
    insert_function(id_2_bigint, CFn.id_2_bigint, conn)
    insert_function(id_2_int, CFn.id_2_int, conn)

    insert_function(jsonb_id_array_2_bigint_array, CFn.jsonb_id_array_2_bigint_array, conn)



import abc
from abc import ABC
from typing import Final

from volsite_postgres_common.db.CC import CC
from volsite_postgres_common.db.FnT import FnT
from volsite_postgres_common.fn.function import ASqlFunction, Arg


class APFnUpsertTitle(ASqlFunction, ABC):
    _title_: Final[str] = '_title_'

    @abc.abstractmethod
    def table_title(self) -> str:  # {ScT.mentor.titles}
        raise NotImplementedError

    @abc.abstractmethod
    def col_title_id(self) -> str:  # {C.title_id}
        raise NotImplementedError

    def in_args(self) -> list:
        return [
            Arg(self._title_, FnT.TEXT)
        ]

    def return_type(self) -> FnT:
        return FnT.BIGINT

    def body(self) -> str:
        return (
            f" INSERT INTO {self.table_title()} ("
            f"     {CC.text}, {CC.length}, {CC.built_time}, {CC.last_used_time}) "
            f" VALUES (trim({self._title_}), length(trim({self._title_})), current_timestamp, current_timestamp) "
            f" ON CONFLICT({CC.text}) "
            f"     DO"
            f"         UPDATE SET "
            f"             {CC.used_count} = {self.table_title()}.{CC.used_count} + 1,  "
            f"             {CC.last_used_time} = current_timestamp"
            f" RETURNING {self.col_title_id()};"
        )



from typing import Final

# ====== BFn: Postgres Build in Functions ======


class BFn:
    array_agg: Final[str] = 'array_agg'
    array_append: Final[str] = 'array_append'
    array_length: Final[str] = 'array_length'
    array_to_string: Final[str] = 'array_to_string'

    cardinality: Final[str] = 'cardinality'
    coalesce: Final[str] = 'coalesce'
    concat: Final[str] = 'concat'
    convert_from: Final[str] = 'convert_from'
    crypt: Final[str] = 'crypt'
    current_timestamp: Final[str] = 'current_timestamp'

    decode: Final[str] = 'decode'

    encode: Final[str] = 'encode'
    enum_range: Final[str] = 'enum_range'

    gen_salt: Final[str] = 'gen_salt'

    icount: Final[str] = 'icount'  # number of elements in array

    jsonb_array_length: Final[str] = 'jsonb_array_length'

    lower: Final[str] = 'lower'

    quote_ident: Final[str] = 'quote_ident'

    jsonb_agg: Final[str] = 'jsonb_agg'
    jsonb_array_elements: Final[str] = 'jsonb_array_elements'
    jsonb_array_elements_text: Final[str] = 'jsonb_array_elements_text'
    jsonb_build_array: Final[str] = 'jsonb_build_array'
    jsonb_build_object: Final[str] = 'jsonb_build_object'
    jsonb_object_agg: Final[str] = 'jsonb_object_agg'
    jsonb_object_keys: Final[str] = 'jsonb_object_keys'
    jsonb_strip_nulls: Final[str] = 'jsonb_strip_nulls'

    length: Final[str] = 'length'

    nextval: Final[str] = 'nextval'

    percentile_cont: Final[str] = 'percentile_cont'
    pg_get_serial_sequence: Final[str] = 'pg_get_serial_sequence'

    sha256: Final[str] = 'sha256'
    similarity: Final[str] = 'similarity'
    sort: Final[str] = 'sort'
    sort_asc: Final[str] = 'sort_asc'
    sort_desc: Final[str] = 'sort_desc'
    split_part: Final[str] = 'split_part'
    substring: Final[str] = 'substring'
    sum: Final[str] = 'sum'

    uniq: Final[str] = 'uniq'

    word_similarity: Final[str] = 'word_similarity'



from typing import Final
from volsite_postgres_common.db.CFn import CFn
from volsite_postgres_common.fn.insert import insert_function


# @ref https://stackoverflow.com/questions/39900397/check-if-anyelement-isnull-postgresql
is_null: Final = (
    f" CREATE OR REPLACE FUNCTION "
    f" {CFn.is_null} (anyelement) "
    f" RETURNS BOOLEAN "
    f" AS"
    f" $$"
    f"   SELECT $1 IS NULL"
    f" $$ "
    f" LANGUAGE SQL "
    f" IMMUTABLE;")

# @ref https://stackoverflow.com/questions/2913368/sorting-array-elements
array_sort: Final = (
    f" CREATE OR REPLACE FUNCTION "
    f" {CFn.array_sort} (ANYARRAY) "
    f" RETURNS ANYARRAY "
    f" AS"
    f" $$"
    f"   SELECT ARRAY(SELECT unnest($1) ORDER BY 1)"
    f" $$ "
    f" LANGUAGE SQL "
    f" IMMUTABLE;")

if_null_2_empty_bigint_array: Final = (
    f" CREATE OR REPLACE FUNCTION "
    f" {CFn.if_null_2_empty_bigint_array} ("
    f" IN _input BIGINT[], "
    f" OUT _result BIGINT[]"
    f" ) "
    f" AS"
    f" $$"
    f" BEGIN "
    f"  IF _input IS NULL THEN"
    f"      _result := ARRAY[]::BIGINT[];"
    f"      RETURN;"
    f"  END IF;"
    f"  _result := _input;"
    f" END;"
    f" $$ "
    f" LANGUAGE PlPgSql "
    f" ;")


def insert_util_fn__general(conn):
    insert_function(is_null, CFn.is_null, conn)
    insert_function(array_sort, CFn.array_sort, conn)
    insert_function(if_null_2_empty_bigint_array, CFn.if_null_2_empty_bigint_array, conn)



from typing import Final


# ====== CFn: Common Custom Function ======
class CFn:
    array_sort: Final[str] = 'array_sort'

    bigint_2_id: Final[str] = 'bigint_2_id'

    count_jsonb_keys: Final[str] = 'count_jsonb_keys'

    id_2_bigint: Final[str] = 'id_2_bigint'
    id_2_int: Final[str] = 'id_2_int'
    if_null_2_empty_bigint_array: Final[str] = 'if_null_2_empty_bigint_array'
    int_2_id: Final[str] = 'int_2_id'
    is_null: Final[str] = 'is_null'

    jsonb_add_key_value_elements: Final[str] = 'jsonb_add_key_value_elements'
    jsonb_array_2_bigint_array: Final[str] = 'jsonb_array_2_bigint_array'
    jsonb_array_2_int_array: Final[str] = 'jsonb_array_2_int_array'
    jsonb_array_2_smallint_array: Final[str] = 'jsonb_array_2_smallint_array'
    jsonb_array_2_text_array: Final[str] = 'jsonb_array_2_text_array'
    jsonb_remove_key: Final[str] = 'jsonb_remove_key'

    jsonb_id_array_2_bigint_array: Final[str] = 'jsonb_id_array_2_bigint_array'



from typing import Final

from volsite_postgres_common.db.CC import CC

NULL_DESCRIPTION_ID: Final = 1


def insert_empty_description(table, description_id, cursor):
    cursor.execute(
        f" INSERT INTO {table} "
        f" ("
        f"   {description_id}, "
        f"   {CC.formatted_description}, {CC.plain_description}, {CC.length}, "
        f"   {CC.used_count}, {CC.built_time}, {CC.last_used_time}"
        f" )"
        f" VALUES"
        f" ("
        f"   {NULL_DESCRIPTION_ID}, "
        f"   '', '', 0, "
        f"   0, current_timestamp, current_timestamp"
        f" );")



import abc
from abc import ABC
from typing import Final

from volsite_postgres_common.db.FnT import FnT
from volsite_postgres_common.db.BFn import BFn
from volsite_postgres_common.db.CFn import CFn
from volsite_postgres_common.fn.function import AJsonPlPgSqlFunction, Arg

from volsite_postgres_common.api.CA import CA


class APFnUpsertTitleDescriptionB64(AJsonPlPgSqlFunction, ABC):

    _title_id: Final[str] = '_title_id'
    _description_id: Final[str] = '_description_id'

    _title: Final[str] = '_title'
    _des_plain: Final[str] = '_des_plain'
    _des_format: Final[str] = '_des_format'

    @abc.abstractmethod
    def input_a_title(self) -> str:  # {A.Title}
        raise NotImplementedError

    @abc.abstractmethod
    def output_a_title_id(self) -> str:  # {A.TitleId}
        raise NotImplementedError

    @abc.abstractmethod
    def pfn_title___upsert(self) -> str:  # {PFn_title___upsert}
        raise NotImplementedError

    @abc.abstractmethod
    def pfn_description___upsert(self) -> str:  # {PFn_description___upsert}
        raise NotImplementedError

    def declare(self) -> list:
        return [
            Arg(self._title_id, FnT.BIGINT),
            Arg(self._description_id, FnT.BIGINT),

            Arg(self._title, FnT.TEXT),
            Arg(self._des_plain, FnT.TEXT),
            Arg(self._des_format, FnT.TEXT)
        ]

    def body(self) -> str:
        return (
            f" {self._title} := {BFn.convert_from}( "
            f"      {BFn.decode}( ({Arg.input}->>'{self.input_a_title()}')::{FnT.TEXT.value}, 'base64'), 'UTF-8');"
            f" {self._des_plain} := {BFn.convert_from}( "
            f"      {BFn.decode}( ({Arg.input}->>'{CA.DescriptionPlain}')::{FnT.TEXT.value}, 'base64'), 'UTF-8');"
            f" {self._des_format} := {BFn.convert_from}( "
            f"      {BFn.decode}( ({Arg.input}->>'{CA.DescriptionFormatted}')::{FnT.TEXT.value}, 'base64'), 'UTF-8');"
            f""
            f" {self._title_id} := {self.pfn_title___upsert()}({self._title}); "
            f" {self._description_id} := {self.pfn_description___upsert()}({self._des_format}, {self._des_plain}); "
            f" {Arg.result} := {BFn.jsonb_build_object}("
            f"       '{self.output_a_title_id()}', {CFn.bigint_2_id}({self._title_id}),"
            f"       '{CA.DescriptionId}', {CFn.bigint_2_id}({self._description_id}) "
            f"   );"
        )



from typing import Final

from volsite_postgres_common.db.CC import CC

NULL_TEXT_ID: Final = 1
DEFAULT_NOTE_TYPE_NAME_ID: Final = 10


def insert_empty_text(table, text_id, cursor):
    cursor.execute(
        f" INSERT INTO {table} ("
        f"     {text_id}, "
        f"     {CC.text}, {CC.length}, "
        f"     {CC.used_count}, {CC.built_time}, {CC.last_used_time} )"
        f" VALUES ("
        f"     {NULL_TEXT_ID}, "
        f"     '', 0,"
        # f"     '$$$sys_null$$$', 0,"
        f"     0, current_timestamp, current_timestamp);")

    cursor.execute(
        f" INSERT INTO {table} ("
        f"     {text_id}, "
        f"     {CC.text}, {CC.length}, "
        f"     {CC.used_count}, {CC.built_time}, {CC.last_used_time} )"
        f" VALUES ("
        f"     {DEFAULT_NOTE_TYPE_NAME_ID}, "
        f"     '$$@$$_NoteType', 0,"
        # f"     '$$$sys_null$$$', 0,"
        f"     0, current_timestamp, current_timestamp);")



from volsite_postgres_common.db.CC import CC
from volsite_postgres_common.table.series import set_max_id_for_serial_column
from volsite_postgres_common.table.text import insert_empty_text


# @note PostgreSQL treats NULL as distinct value
def create_table_titles(conn, sc: str, table: str, title_id: str, add_trgm_index: bool = True):
    sc_table = f'"{sc}"."{table}"'
    cursor = conn.cursor()
    cursor.execute(
        f" CREATE TABLE IF NOT EXISTS "
        f" {sc_table} ("
        f"     {title_id}            BIGSERIAL      PRIMARY KEY, "  # 8bytes 
        f"     {CC.text}              TEXT           NOT NULL, "
        f"     {CC.length}            SMALLINT       NOT NULL, "
        f"     {CC.used_count}        INTEGER        NOT NULL DEFAULT 1, "
        f"     {CC.built_time}        TIMESTAMP      NOT NULL, "
        f"     {CC.last_used_time}    TIMESTAMP      NOT NULL );")

    if add_trgm_index:
        cursor.execute(f"CREATE INDEX trgm_idx ON {sc_table} USING GIN ({CC.text} gin_trgm_ops);")

    cursor.execute(
        f" CREATE UNIQUE INDEX {sc}_{table}_{CC.text} ON "
        f"     {sc_table}("
        f"         {CC.text});")

    insert_empty_text(sc_table, title_id, cursor)

    set_max_id_for_serial_column(sc, table, title_id, cursor)
    conn.commit()



from typing import Final


# ====== CA: Common Attribute ======

class CA:
    Api: Final[str] = "api"
    Auth: Final[str] = "aut"
    Ascending: Final[str] = "asc"  # in SQL

    Check: Final[str] = "chk"
    Click: Final[str] = "clk"
    Code: Final[str] = "cd"
    Collect: Final[str] = "clt"
    Collected: Final[str] = "cltd"
    Color: Final[str] = "col"
    Count: Final[str] = "cnt"
    Create: Final[str] = "ct"
    CreateTime: Final[str] = "crt"

    Data: Final[str] = "dt"
    Descending: Final[str] = "desc"  # in SQL
    Description: Final[str] = "des"
    DescriptionFormatted: Final[str] = "desf"
    DescriptionId: Final[str] = "desid"
    DescriptionPlain: Final[str] = "desp"

    Finished: Final[str] = "fied"

    Head: Final[str] = "hd"
    HeadVersion: Final[str] = "hv"

    Id: Final[str] = "id"
    Info: Final[str] = "inf"

    List: Final[str] = "lt"
    Login: Final[str] = "lgn"

    Name: Final[str] = "na"
    Note: Final[str] = "nt"

    Password: Final[str] = "pw"
    Public: Final[str] = "pb"

    Request: Final[str] = "rqt"
    Result: Final[str] = "ret"

    Signup: Final[str] = "sgu"
    SortBy: Final[str] = "stby"
    SortDirection: Final[str] = "stdr"
    Stars: Final[str] = "strs"
    Start: Final[str] = "str"
    State: Final[str] = "ste"

    Text: Final[str] = "txt"
    Time: Final[str] = "tm"
    TimeZone: Final[str] = "tmz"
    Title: Final[str] = "tt"
    TitleId: Final[str] = "ttid"
    Type: Final[str] = "tp"

    Update: Final[str] = "ud"
    UpdateTime: Final[str] = "udt"
    Used: Final[str] = "used"
    User: Final[str] = "usr"
    UserId: Final[str] = "uid"

    Version: Final[str] = "vn"
    VersionCreate: Final[str] = "vct"
    VersionId: Final[str] = "vid"
    Visited: Final[str] = "vst"
    Visiting: Final[str] = "vstng"

    Word: Final[str] = "w"

    ___Error___: Final[str] = "err"
